(* finding cmr10 font *)
(*
let cmr10 = Font.find "cmr10" 600;;
let cmr_A = Font.find_char_def cmr10 (int_of_char 'A');;
Printf.printf "cmr10 A: %d, %d, %d\n"
  cmr_A.Font.code cmr_A.Font.width cmr_A.Font.height;;
*)

(* device fonts *)
module DFont = Devfont.Make(Grdev);;

let pt = 65536;;
let fpt = float pt;;

type glue_setting =
  | Stretching of float
  | Shrinking of float
  | Filling of float
  | Rigid;;

let print_glue_setting = function
  | Stretching r -> Printf.printf "Stretching %g" r
  | Shrinking r -> Printf.printf "Shrinking %g" r
  | Filling r -> Printf.printf "Filling %g" r
  | Rigid -> print_string "Rigid";;

(* drawing state *)
type state = {
  mutable x:	int;
  mutable y:	int;
  mutable g:	glue_setting;
};;

(* FONT *)
type font = {
  name:		string;
  mtable:	(int * int) Table.t;
  gtable:	Grdev.glyph Table.t;
};;

let dummy_mtable = Table.make (fun _ -> raise Not_found);;
let dummy_gtable = Table.make (fun _ -> raise Not_found);;

let make_font name dpi =
  let mtable =
    try DFont.find_metrics name dpi
    with Not_found -> dummy_mtable in
  let gtable =
    try DFont.find_glyphs name dpi
    with Not_found -> dummy_gtable in
  { name = name;
    mtable = mtable;
    gtable = gtable; };;

(* debugging the font parameters for a char *)
let inspect font code =
  let i = int_of_char code in
  let dev_glyph = Table.get font.gtable i
  and dx, dy = Table.get font.mtable i in
  let glyph = Grdev.get_glyph dev_glyph in
  print_string "character "; print_char code; print_string ":"; print_newline ();
  print_string "  width   = ";   print_int glyph.Glyph.width;   print_string ", ";
  print_string "  height  = ";  print_int glyph.Glyph.height;  print_newline ();
  print_string "  hoffset = "; print_int glyph.Glyph.hoffset; print_string ", ";
  print_string "  voffset = "; print_int glyph.Glyph.voffset; print_newline ();
  print_string "  dx = "; print_int dx; print_string " (";
  print_int (Misc.round (float dx /. fpt)); print_char ')';
  print_newline ();;

(* puts the character in the given font on the screen, updating the drawing state *)
let put_char st font code =
  let i = int_of_char code in
  let dev_glyph = Table.get font.gtable i
  and dx, dy = Table.get font.mtable i
  and x = Misc.round (float st.x /. fpt)
  and y = Misc.round (float st.y /. fpt) in
  Grdev.draw_glyph dev_glyph x y;
  st.x <- st.x + dx;
  st.y <- st.y + dy;;

(* GLUE (elastic dimension) *)
type glue = { nat: int; shr: int; str: int; fil: int; };;

let zero_glue = { nat = 0; shr = 0; str = 0; fil = 0; };;

(* adds two glues *)
let add_glue g h = {
  nat = g.nat + h.nat; shr = g.shr + h.shr;
  str = g.str + h.str; fil = g.fil + h.fil;
};;

let print_glue g =
  print_string "{ nat = "; print_int g.nat;
  print_string ", shr = "; print_int g.shr;
  print_string ", str = "; print_int g.str;
  print_string ", fil = "; print_int g.fil;
  print_string " }";;

let glue_bounds g max =
  let lower = g.nat - g.shr
  and upper =
    if g.fil <> 0 then max
    else min (g.nat + g.str) max in
  (lower, upper);;

let set_glue g = function
  | Stretching r -> g.nat + Misc.round (float g.str *. r)
  | Shrinking r ->  g.nat - Misc.round (float g.shr *. r)
  | Filling r ->    g.nat + Misc.round (float g.fil *. r)
  | Rigid ->        g.nat;;

let setting_for d g =
  if d < g.nat then
    if g.shr <> 0 then Shrinking (float (g.nat - d) /. float g.shr)
    else Rigid
  else if d > g.nat then
    if g.fil <> 0 then Filling (float (d - g.nat) /. float g.fil)
    else if g.str <> 0 then Stretching (float (d - g.nat) /. float g.str)
    else Rigid
  else Rigid;;

(* NODE *)
type node =
  | Char of char * font
  | Skip of glue
  (* | Pen of int *)
  | Break;;

(* prints the node mnemonicly on stdout *)
let print_node = function
  | Char (code, font) ->
    print_string "Char ("; print_char code;
    print_string ", "; print_string font.name;
    print_char ')'
  | Skip g ->
    print_string "Skip "; print_glue g
(*
  | Pen p ->
    print_string "Pen ("; print_int p; print_char ')'
*)
  | Break ->
    print_string "Break"
  ;;

(* puts the node on the screen, updating the drawing state *)
let put_node st = function
  | Char (code, font) -> put_char st font code
  | Skip g -> st.x <- st.x + set_glue g st.g
  (* | Pen p -> () *)
  | Break -> st.x <- 0; st.y <- st.y + 36 * pt;;

(* adds a glue and the (elastic) width of a node *)
let add_glue_width g = function
  | Char (code, font) ->
    let i = int_of_char code in
    let dx, _ = Table.get font.mtable (int_of_char code) in
    { g with nat = g.nat + dx }
  | Skip h -> add_glue g h
  | Break -> g;;

(* auxiliary buffer for the nodes *)
type 'a list_buffer = {
  mutable nodes: node list;
  mutable count: int;
};;

(*
 * reads the text from a file and returns the corresponding array of nodes.
 * font:	the font used for the non-space characters
 * file:	name of the file
 *)
let read_file font file =
  let input = open_in file
  and space = Skip { nat = 10 * pt; shr = 10 * pt / 3; str = 5 * pt; fil = 0; }
  and fil = Skip { nat = 0; shr = 0; str = 0; fil = pt; }
  and buff = { nodes = []; count = 0; } in
  let make_node = function
    | ' ' | '\t' | '\n' -> space
    | code -> Char (code, font)
  and add node = buff.nodes <- node :: buff.nodes; buff.count <- succ buff.count in
  begin try
    let fresh = ref true in
    while true do
      let line = input_line input in
      if String.length line > 0 then begin
	if !fresh then fresh := false else add space;
	String.iter (function c -> add (make_node c)) line
      end else begin add fil; add Break; fresh := true end
    done
  with End_of_file -> add fil; close_in input
  end;
  let text = Array.make buff.count Break
  and put (t, i) node = t.(i) <- node; (t, i - 1) in
  ignore (List.fold_left put (text, buff.count - 1) buff.nodes);
  text;;

(* representation of a cost *)
type demerit =
  | Dem of int
  | NoWay;;

let zero_demerit = Dem 0;;

let inf_bad = 10000;;
let inf_bad_dem = inf_bad * inf_bad;;

let no_way = function
  | Dem _ -> false
  | NoWay -> true;;

let dem_count = function
  | Dem d -> d
  | NoWay -> raise (Failure "No way");;

let add_demerits x y = match x, y with
  | Dem dx, Dem dy ->
    let d = dx + dy in
    if d < inf_bad_dem then Dem d else NoWay
  | _ -> NoWay;;

let print_demerit = function
  | Dem d -> print_int d
  | NoWay -> print_string "NoWay";;

(* badness of stretching/shrinking by dim compared to lim *)
let rec badness_of dim lim =
  let ratio r =
    if r > 1290 then inf_bad
    else (r * r * r + 0400000) / 01000000 in
  if dim = 0 then 0
  else if dim < 0 then badness_of (- dim) (- lim)
  else if lim <= 0 then inf_bad
  else if dim <= 7230584 || lim >= 1663497 then
    Misc.round (float dim *. 297.0 /. float lim)
(*
  else if dim <= 7230584 then ratio (dim * 297 / lim)
  else if lim >= 1663497 then ratio (dim / (lim / 297))
*)
  else inf_bad;;

let demerits_of badness =
  if badness < inf_bad
  then Dem (badness * badness)
  else NoWay;;

(* element of a breaking path *)
type break = {
  index:	int;		(* index into array of nodes (line to be broken) *)
  serial:	int;		(* serial number of the break at the given index *)
  next_idx:	int;		(* starting index of the next line if the break is taken *)
  width:	glue;		(* elastic width of the resulting line *)
  dy:		glue;		(* soft vertical position of the next line *)
  min_dem:	demerit;	(* minimal demerits known (estimated so far) *)
  max_dem:	demerit;	(* maximal demerits known (estimated so far) *)
  ideal:	int;
};;

let zero_break = {
  index = 0; serial = 0; next_idx = 0;
  width = zero_glue; dy = zero_glue;
  min_dem = zero_demerit; max_dem = zero_demerit;
  ideal = 0;
};;

(*
 * calculates the minimum and maximum costs of a braking path
 * nodes:	array of nodes to be broken (whole paragraph or text)
 * index:	the position of this potential break
 * path:	breaking path to be prolonged by this potential break
 * returns (minimal demerits, maximal demerits, line too long)
 *)
let path_cost (lower, upper) nodes index path =
  let brk = List.hd path in
  let width = Array.fold_left add_glue_width zero_glue
  		(Array.sub nodes brk.next_idx (index - brk.next_idx)) in
  let badness_for expected =
    let diff = width.nat - expected in
    if diff < 0 then
      if width.fil <> 0 then (0, false)
      else (badness_of (- diff) width.str, false)
    else if diff > width.shr then (inf_bad, true)
    else (badness_of diff width.shr, false) in
  (* if lower = upper then don't calculate both *)
  let ub, utl = badness_for upper
  and lb, ltl = badness_for lower in
  let too_long = utl && ltl in
  if width.nat < lower then
    (demerits_of lb, demerits_of ub, too_long, width, lower)
  else if width.nat > upper then
    (demerits_of ub, demerits_of lb, too_long, width, upper)
  else (zero_demerit, demerits_of (max lb ub), false, width, width.nat);;

(*
 * trying the next breaking position
 * nodes:	array of nodes to be broken (whole paragraph or text)
 * active:	the feasible breaking paths so far
 * index:	the position of this potential break
 * skip:	how many nodes have to discarded if the break is taken
 * forced:	is it a hard break?
 * returns an updated list of the feasible paths
 *)
let try_break desired vert_step nodes active index next_idx forced =
  let consider path (updated, serial) =
    let brk = List.hd path in
    if index <= brk.next_idx then path :: updated, serial
    else begin
      let bounds = desired brk.dy in
      let md, xd, too_long, width, ideal = path_cost bounds nodes index path in
      if too_long then updated, serial else
	let min_dem = add_demerits md brk.min_dem
	and max_dem = add_demerits xd brk.max_dem in
	if no_way min_dem then
	  (if forced then updated else path :: updated), serial
	else let feasible = {
	  index = index; serial = serial;
	  next_idx = next_idx; width = width;
	  dy = add_glue brk.dy (vert_step path);
	  min_dem = min_dem; max_dem = max_dem;
	  ideal = ideal;
	} in
	(*D*) Printf.printf "\t@@%d.%d -> @@%d.%d\t" index serial brk.index brk.serial;
	(*D*) print_string " md="; print_demerit md;
	(*D*) print_string " xd="; print_demerit xd;
	(*D*) print_string " min_dem="; print_demerit min_dem;
	(*D*) print_string " max_dem="; print_demerit max_dem;
	(*D*) print_newline ();
	( if forced then (feasible :: path) :: updated
	  else (feasible :: path) :: path :: updated ), succ serial
    end in
  List.fold_right consider active ([], 0);;

(*
 * breaking the list of nodes into lines
 *)
let break_text desired vert_step nodes =
  let len = Array.length nodes in
  let rec only_next i =
    if i < len then
      match nodes.(i) with
	| Skip _ | Break -> only_next (succ i)
	| Char _ -> i
    else i in
  let rec next_step i =
    if i < len then
      match nodes.(i) with
	| Skip _ -> next_step (succ i)
	| Break -> (only_next (succ i), i)
	| Char _ -> (i, i)
    else (i, i) in
  let active = ref [[zero_break]]
  and count = ref 0
  (*D*) and start_chars = ref true
  and idx = ref 0 in
  (*D*) let finish_chars () =
    if (not !start_chars) then begin print_newline (); start_chars := true; end in
  let try_and_go forced =
    let next_idx, step = next_step (succ !idx) in
    let act, num = try_break desired vert_step nodes !active !idx next_idx forced in
    active := act; count := !count + num; idx := step in
  while !idx < len do
    match nodes.(!idx) with
      | Skip _ ->
        (*D*) finish_chars ();
        (* (*D*) Printf.printf "\t** Trying to break at Skip %d\n" !idx; *)
        try_and_go false
      | Break ->
        (*D*) finish_chars ();
        (* (*D*) Printf.printf "\t** Forcing break at Break %d\n" !idx; *)
        try_and_go true
      | Char (code, _) ->
        (*D*) print_char code; start_chars := false;
	incr idx;
  done;
  (*D*) finish_chars ();
  (* (*D*) Printf.printf "\t** Forcing break at the end %d\n" len; *)
  let act, num = try_break desired vert_step nodes !active len len true in
  active := act; count := !count + num;
  (*D*) Printf.printf "%d Breaks used\n" !count;
  let min_path x y =
    let xb = List.hd x
    and yb = List.hd y in
    if xb.min_dem < yb.min_dem then x else y in
  match !active with
  | first :: rest ->
    let best = List.fold_left min_path first rest in
    let brk = List.hd best in
    (*D*) Printf.printf "best %d.%d\n" brk.index brk.serial;
    best
  | [] ->
    (*D*) print_string "no path found :-(\n";
    [];;

let rec put_par y setting get_width nodes = function
  | curr :: prev :: rest ->
    put_par y setting get_width nodes (prev :: rest);
    let line = Array.sub nodes prev.next_idx (curr.index - prev.next_idx)
    and h = set_glue prev.dy setting in
    (* let st = { x = 0; y = y + h; g = setting_for (get_width h) curr.width; } in *)
    let st = { x = 0; y = y + h; g = setting_for curr.ideal curr.width; } in
    Array.iter (put_node st) line;
  | _ -> ();;


(* M A I N *)

if Array.length Sys.argv < 2
then failwith "Missing file argument";;

(* initializing graphical device *)
let width, height = Grdev.open_dev " 1200x900";;
Graphics.set_window_title "Dyntyp";;
Graphics.auto_synchronize true;;

(* drawing (mostly) *)
let state = { x = 0; y = 100 * pt; g = Rigid; };;
let cmr10 = make_font "cmr10" 200.0;;
(* String.iter (inspect cmr10) "abcdefghijklmnopqrstuvwxyz"; *)
(* String.iter (put_char state cmr10) "ahoj";; *)
(* Grdev.fill_rect 0 100 width 5;; *)

(*
 * reading the file given on the command line
 * returning the corresponding list of nodes
 *)
let text = read_file cmr10 Sys.argv.(1);;

(* Array.iter (function c -> print_node c; print_newline ()) text;; *)
(* Array.iter (put_node state) text;; *)

let vert_step _ = { zero_glue with nat = 36 * pt; fil = pt; };;
let top = 36 * pt;;
let vsize = 800 * pt;;

let hsize d = 600 * pt + d / 3;;

let desired g = 
  let lower, upper = glue_bounds g vsize in
  (hsize lower, hsize upper);;

let par = break_text desired vert_step text;;
let dy = (List.hd par).dy;;
let setting = setting_for vsize dy;;

(*
Printf.printf "vsize = %d\n" vsize;;
print_string "dy = "; print_glue dy; print_newline ();;
print_string "setting = "; print_glue_setting setting; print_newline ();;
*)

put_par top setting hsize text par;;

let my_plot top hsize vsize =
  Graphics.set_color Graphics.red;
  let t = Misc.round (float top /. fpt) in
  let y = ref t
  and h = Graphics.size_y () in
  while !y * pt <= top + vsize do
    let x = Misc.round (float (hsize ((!y - t) * pt)) /. fpt) in
    Graphics.plot x (h - !y);
    y := succ !y
  done;;

my_plot top hsize vsize;;

(* waiting for a key *)

let rec wait_for_key () =
  match Grdev.wait_event () with
    | Grdev.Key c -> c
    | _ -> wait_for_key ()
;;

prerr_endline "Press a key";;
wait_for_key ();;
Grdev.close_dev ();;
