(* finding cmr10 font *)
(*
let cmr10 = Font.find "cmr10" 600;;
let cmr_A = Font.find_char_def cmr10 (int_of_char 'A');;
Printf.printf "cmr10 A: %d, %d, %d\n"
  cmr_A.Font.code cmr_A.Font.width cmr_A.Font.height;;
*)

(* device fonts *)
module DFont = Devfont.Make(Grdev);;

let pt = 65536;;
let fpt = float pt;;

type glue_setting =
  | Stretching of float
  | Shrinking of float
  | Filling of float
  | Rigid;;

let print_glue_setting = function
  | Stretching r -> Printf.printf "Stretching %g" r
  | Shrinking r -> Printf.printf "Shrinking %g" r
  | Filling r -> Printf.printf "Filling %g" r
  | Rigid -> print_string "Rigid";;

(* drawing state *)
type state = {
  mutable x:	int;
  mutable y:	int;
  mutable g:	glue_setting;
};;

(* FONT *)
type font = {
  name:		string;
  mtable:	(int * int) Table.t;
  gtable:	Grdev.glyph Table.t;
};;

let dummy_mtable = Table.make (fun _ -> raise Not_found);;
let dummy_gtable = Table.make (fun _ -> raise Not_found);;

let make_font name dpi =
  let mtable =
    try DFont.find_metrics name dpi
    with Not_found -> dummy_mtable in
  let gtable =
    try DFont.find_glyphs name dpi
    with Not_found -> dummy_gtable in
  { name = name;
    mtable = mtable;
    gtable = gtable; };;

(* debugging the font parameters for a char *)
let inspect font code =
  let i = int_of_char code in
  let dev_glyph = Table.get font.gtable i
  and dx, dy = Table.get font.mtable i in
  let glyph = Grdev.get_glyph dev_glyph in
  print_string "character "; print_char code; print_string ":"; print_newline ();
  print_string "  width   = ";   print_int glyph.Glyph.width;   print_string ", ";
  print_string "  height  = ";  print_int glyph.Glyph.height;  print_newline ();
  print_string "  hoffset = "; print_int glyph.Glyph.hoffset; print_string ", ";
  print_string "  voffset = "; print_int glyph.Glyph.voffset; print_newline ();
  print_string "  dx = "; print_int dx; print_string " (";
  print_int (Misc.round (float dx /. fpt)); print_char ')';
  print_newline ();;

(* puts the character in the given font on the screen, updating the drawing state *)
let put_char st font code =
  let i = int_of_char code in
  let dev_glyph = Table.get font.gtable i
  and dx, dy = Table.get font.mtable i
  and x = Misc.round (float st.x /. fpt)
  and y = Misc.round (float st.y /. fpt) in
  Grdev.draw_glyph dev_glyph x y;
  st.x <- st.x + dx;
  st.y <- st.y + dy;;

(* GLUE (elastic dimension) *)
type glue = { nat: int; shr: int; str: int; fil: int; };;

let zero_glue = { nat = 0; shr = 0; str = 0; fil = 0; };;

(* adds two glues *)
let add_glue g h = {
  nat = g.nat + h.nat; shr = g.shr + h.shr;
  str = g.str + h.str; fil = g.fil + h.fil;
};;

let print_glue g =
  print_string "{ nat = "; print_int g.nat;
  print_string ", shr = "; print_int g.shr;
  print_string ", str = "; print_int g.str;
  print_string ", fil = "; print_int g.fil;
  print_string " }";;

let glue_bounds g max =
  let lower = g.nat - g.shr
  and upper =
    if g.fil <> 0 then max
    else min (g.nat + g.str) max in
  (lower, upper);;

let set_glue g = function
  | Stretching r -> g.nat + Misc.round (float g.str *. r)
  | Shrinking r ->  g.nat - Misc.round (float g.shr *. r)
  | Filling r ->    g.nat + Misc.round (float g.fil *. r)
  | Rigid ->        g.nat;;

let setting_for d g =
  if d < g.nat then
    if g.shr <> 0 then Shrinking (float (g.nat - d) /. float g.shr)
    else Rigid
  else if d > g.nat then
    if g.fil <> 0 then Filling (float (d - g.nat) /. float g.fil)
    else if g.str <> 0 then Stretching (float (d - g.nat) /. float g.str)
    else Rigid
  else Rigid;;

(* NODE *)
type node =
  | Char of char * font
  | Skip of glue
  (* | Pen of int *)
  | Break;;

(* prints the node mnemonicly on stdout *)
let print_node = function
  | Char (code, font) ->
    print_string "Char ("; print_char code;
    print_string ", "; print_string font.name;
    print_char ')'
  | Skip g ->
    print_string "Skip "; print_glue g
(*
  | Pen p ->
    print_string "Pen ("; print_int p; print_char ')'
*)
  | Break ->
    print_string "Break"
  ;;

(* puts the node on the screen, updating the drawing state *)
let put_node st = function
  | Char (code, font) -> put_char st font code
  | Skip g -> st.x <- st.x + set_glue g st.g
  (* | Pen p -> () *)
  | Break -> st.x <- 0; st.y <- st.y + 36 * pt;;

(* adds a glue and the (elastic) width of a node *)
let add_glue_width g = function
  | Char (code, font) ->
    let i = int_of_char code in
    let dx, _ = Table.get font.mtable (int_of_char code) in
    { g with nat = g.nat + dx }
  | Skip h -> add_glue g h
  | Break -> g;;

(* auxiliary buffer for the nodes *)
type 'a list_buffer = {
  mutable nodes: node list;
  mutable count: int;
};;

(*
 * reads the text from a file and returns the corresponding array of nodes.
 * font:	the font used for the non-space characters
 * file:	name of the file
 *)
let read_file font file =
  let input = open_in file
  and space = Skip { nat = 10 * pt; shr = 10 * pt / 3; str = 5 * pt; fil = 0; }
  and fil = Skip { nat = 0; shr = 0; str = 0; fil = pt; }
  and buff = { nodes = []; count = 0; } in
  let make_node = function
    | ' ' | '\t' | '\n' -> space
    | code -> Char (code, font)
  and add node = buff.nodes <- node :: buff.nodes; buff.count <- succ buff.count in
  begin try
    let fresh = ref true in
    while true do
      let line = input_line input in
      if String.length line > 0 then begin
	if !fresh then fresh := false else add space;
	String.iter (function c -> add (make_node c)) line
      end else begin add fil; add Break; fresh := true end
    done
  with End_of_file -> add fil; close_in input
  end;
  let text = Array.make buff.count Break
  and put (t, i) node = t.(i) <- node; (t, i - 1) in
  ignore (List.fold_left put (text, buff.count - 1) buff.nodes);
  text;;

(* representation of a cost *)
type demerit =
  | Dem of int
  | NoWay;;

let zero_demerit = Dem 0;;

let inf_bad = 10000;;
let inf_bad_dem = inf_bad * inf_bad;;

let no_way = function
  | Dem _ -> false
  | NoWay -> true;;

let dem_count = function
  | Dem d -> d
  | NoWay -> raise (Failure "No way");;

let add_demerits x y = match x, y with
  | Dem dx, Dem dy ->
    let d = dx + dy in
    if d < inf_bad_dem then Dem d else NoWay
  | _ -> NoWay;;

let compare_demerits x y = match x, y with
  | Dem dx, Dem dy -> compare dx dy
  | Dem dx, NoWay -> -1
  | NoWay, Dem dy -> 1
  | NoWay, NoWay -> 0;;

let min_demerit x y = match x, y with
  | Dem dx, Dem dy -> Dem (min dx dy)
  | Dem dx, NoWay -> Dem dx
  | NoWay, Dem dy -> Dem dy
  | _ -> NoWay;;

let max_demerit x y = match x, y with
  | Dem dx, Dem dy -> Dem (max dx dy)
  | _ -> NoWay;;

let print_demerit = function
  | Dem d -> print_int d
  | NoWay -> print_string "NoWay";;

(* badness of stretching/shrinking by dim compared to lim *)
let rec badness_of dim lim =
  let ratio r =
    if r > 1290 then inf_bad
    else (r * r * r + 0400000) / 01000000 in
  if dim = 0 then 0
  else if dim < 0 then badness_of (- dim) (- lim)
  else if lim <= 0 then inf_bad
  else if dim <= 7230584 || lim >= 1663497 then
    Misc.round (float dim *. 297.0 /. float lim)
(*
  else if dim <= 7230584 then ratio (dim * 297 / lim)
  else if lim >= 1663497 then ratio (dim / (lim / 297))
*)
  else inf_bad;;

let demerits_of badness =
  if badness < inf_bad
  then Dem (badness * badness)
  else NoWay;;

(* element of a breaking graph *)
type break = {
  index:	int;		(* index into array of nodes (line to be broken) *)
  serial:	int;		(* serial number of the break at the given index *)
  next_idx:	int;		(* starting index of the next line if the break is taken *)
  vpos:		glue;		(* soft vertical position of the line ending here*)
  min_dem:	demerit;	(* minimal demerits known (estimated so far) *)
  max_dem:	demerit;	(* maximal demerits known (estimated so far) *)
  prevs:	break list;	(* list of the preceding breaks *)
};;

let zero_break = {
  index = 0; serial = 0; next_idx = 0; vpos = zero_glue;
  min_dem = zero_demerit; max_dem = zero_demerit;
  prevs = [];
};;

let compare_breaks a b = compare a.vpos b.vpos;;

let print_break brk =
  Printf.printf "{ %3d.%03d, %3d, " brk.index brk.serial brk.next_idx;
  print_glue brk.vpos;
  print_string ", "; print_demerit brk.min_dem;
  print_string ", "; print_demerit brk.max_dem;
  print_string ", [ ";
  List.iter (fun b -> Printf.printf "%3d.%03d; " b.index b.serial) brk.prevs; 
  print_string "] }";;

let badness_for width expected =
  let diff = width.nat - expected in
  if diff < 0 then
    if width.fil <> 0 then (0, false)
    else (badness_of (- diff) width.str, false)
  else if diff > width.shr then (inf_bad, true)
  else (badness_of diff width.shr, false);;

(*
 * calculates the minimum and maximum costs of a braking path
 * nodes:	array of nodes to be broken (whole paragraph or text)
 * index:	the position of this potential break
 * brk:		break to be followed by this potential one
 * returns (minimal demerits, maximal demerits, line too long)
 *)
let break_cost (lower, upper) nodes index brk =
  let width = Array.fold_left add_glue_width zero_glue
  		(Array.sub nodes brk.next_idx (index - brk.next_idx)) in
  (* if lower = upper then don't calculate both *)
  let ub, utl = badness_for width upper
  and lb, ltl = badness_for width lower in
  let too_long = utl && ltl in
  if width.nat < lower then
    (demerits_of lb, demerits_of ub, too_long)
  else if width.nat > upper then
    (demerits_of ub, demerits_of lb, too_long)
  else (zero_demerit, demerits_of (max lb ub), false);;

type trying_context = {
  c_serial:	int;
  c_vpos:	glue;
  c_min_dem:	demerit;
  c_max_dem:	demerit;
  c_prevs:	break list;
  news:		break list;
  olds:		break list;
};;

(*
 * trying the next breaking position
 * nodes:	array of nodes to be broken (whole paragraph or text)
 * active:	the feasible breaks so far
 * index:	the position of this potential break
 * skip:	how many nodes have to discarded if the break is taken
 * returns an updated list of the feasible paths
 *)
let try_break desired vert_step nodes active index next_idx forced =
  let closed_pendings context =
    if context.c_prevs = [] then context
    else let brk = {
      index = index; serial = context.c_serial;
      next_idx = next_idx; vpos = context.c_vpos;
      min_dem = context.c_min_dem;
      max_dem = context.c_max_dem;
      prevs = context.c_prevs;
    } in {
      context with
      c_serial = succ context.c_serial;
      c_prevs = []; news = brk :: context.news;
    } in
  let consider context brk =
    if index <= brk.next_idx then { context with olds = brk :: context.olds }
    else begin
      let vpos = add_glue brk.vpos (vert_step brk) in
      let bounds = desired vpos in
      let md, xd, too_long = break_cost bounds nodes index brk in
      if too_long then context else
	let min_dem = add_demerits md brk.min_dem
	and max_dem = add_demerits xd brk.max_dem in
	if no_way min_dem then { context with olds = brk :: context.olds }
	else begin
	  (*D*) Printf.printf "\t@@%3d.%03d -> @@%3d.%03d\t" index context.c_serial brk.index brk.serial;
	  (*D*) print_string " md="; print_demerit md;
	  (*D*) print_string " xd="; print_demerit xd;
	  (*D*) print_string " min_dem="; print_demerit min_dem;
	  (*D*) print_string " max_dem="; print_demerit max_dem;
	  (*D*) print_string " vpos="; print_glue vpos;
	  (*D*) print_newline ();
	  if vpos = context.c_vpos then {
	    context with
	    c_min_dem = min_demerit context.c_min_dem min_dem;
	    c_max_dem = max_demerit context.c_max_dem max_dem;
	    c_prevs = brk :: context.c_prevs;
	    olds = brk :: context.olds;
	  } else let context = closed_pendings context in {
	    context with
	    c_vpos = vpos;
	    c_min_dem = min_dem;
	    c_max_dem = max_dem;
	    c_prevs = brk :: context.c_prevs;
	    olds = brk :: context.olds;
	  }
	end
    end in
  let context = {
    c_serial = 0; c_vpos = zero_glue;
    c_min_dem = zero_demerit; c_max_dem = zero_demerit;
    c_prevs = []; news = []; olds = [];
  } in
  let result = closed_pendings (List.fold_left consider context active) in
  let news = List.sort compare_breaks result.news in
  (if forced then news else List.merge compare_breaks news (List.rev result.olds)),
  result.c_serial;;

(*
 * breaking the list of nodes into lines
 *)
let break_text desired vert_step nodes =
  let len = Array.length nodes in
  let rec only_next i =
    if i < len then
      match nodes.(i) with
	| Skip _ | Break -> only_next (succ i)
	| Char _ -> i
    else i in
  let rec next_step i =
    if i < len then
      match nodes.(i) with
	| Skip _ -> next_step (succ i)
	| Break -> (only_next (succ i), i)
	| Char _ -> (i, i)
    else (i, i) in
  let active = ref [zero_break]
  and count = ref 0
  (*D*) and start_chars = ref true
  and idx = ref 0 in
  (*D*) let finish_chars () =
    if (not !start_chars) then begin print_newline (); start_chars := true; end in
  let trace_breaks breaks index =
    let trace brk =
      if brk.index <> index then print_char '\t';
      print_string "\t\t"; print_break brk;
      print_newline () in
    List.iter trace breaks in
  let try_and_go forced =
    let next_idx, step = next_step (succ !idx) in
    let act, num = try_break desired vert_step nodes !active !idx next_idx forced in
    active := act; count := !count + num;
    trace_breaks !active !idx;
    idx := step in
  trace_breaks !active !idx;
  while !idx < len do
    match nodes.(!idx) with
      | Skip _ ->
        (*D*) finish_chars ();
        (* (*D*) Printf.printf "\t** Trying to break at Skip %3d\n" !idx; *)
        try_and_go false
      | Break ->
        (*D*) finish_chars ();
        (* (*D*) Printf.printf "\t** Forcing break at Break %3d\n" !idx; *)
        try_and_go true
      | Char (code, _) ->
        (*D*) print_char code; start_chars := false;
	incr idx;
  done;
  (*D*) finish_chars ();
  (* (*D*) Printf.printf "\t** Forcing break at the end %3d\n" len; *)
  let act, num = try_break desired vert_step nodes !active len len true in
  active := act; count := !count + num;
  trace_breaks !active !idx;
  (*D*) Printf.printf "%d Breaks used\n" !count;
  !active;;

let same_break b1 b2 = b1.index = b2.index && b1.serial = b2.serial;;

let rec replace_or_drop curr paths =
  let path, c1, _ = curr in
  let b1 = List.hd path
  and p1 = List.tl path in
  let rec remove_from = function
    | (((b2 :: p2, c2, _) as head) :: rest) ->
      if same_break b1 b2 then rest
      else head :: remove_from rest
    | _ -> [] in
  match paths with
    | (((b2 :: p2, c2, _) as head) :: rest) ->
      if c1 < c2 then curr :: remove_from paths		(* replace *)
      else if same_break b1 b2 then paths		(* drop *)
      else head :: replace_or_drop curr rest		(* try further *)
    | _ -> [curr];;

let relax_edges nodes get_width (path, c, setting) active breaks = 
  let curr = List.hd path
  and rest = List.tl path in
  let relax active prev =
    let width = Array.fold_left add_glue_width zero_glue
		  (Array.sub nodes prev.next_idx (curr.index - prev.next_idx))
    and expected = get_width (set_glue curr.vpos setting) in
    let badness, too_long = badness_for width expected in
    if too_long then active
    else let dem = add_demerits (demerits_of badness) (Dem c) in
      if no_way dem then active
      else let cost = dem_count dem in
        replace_or_drop (prev :: path, cost, setting) active in
  List.fold_left relax active breaks;;

let best_par nodes get_width vsize breaks =
  let trac (path, cost, setting) =
    Printf.printf "%10d\t" cost;
    List.iter (fun brk -> Printf.printf "%3d.%03d " brk.index brk.serial) path;
    print_newline () in
  let trace active = List.iter (fun x -> print_string "\t"; trac x) active  in
  let init active brk =
    let setting = setting_for vsize brk.vpos in
    relax_edges nodes get_width ([brk], 0, setting) active brk.prevs in
  let active = ref (List.fold_left init [] breaks)
  and best = ref []
  and sett = ref Rigid in
  (*
  print_string "searching initialized\n";
  trace !active;
  *)
  while !active <> [] && !best = [] do
    let ((path, c, setting) as first) = List.hd !active in
    (* print_string "choosing "; trac first; print_newline (); *)
    let brk = List.hd path in
    if same_break brk zero_break then begin best := path; sett := setting end
    else active := relax_edges nodes get_width first (List.tl !active) brk.prevs;
    (* trace !active *)
  done;
  (!best, !sett);;

let rec put_par y setting get_indent get_width nodes = function
  | curr :: next :: rest ->
    let line = Array.sub nodes curr.next_idx (next.index - curr.next_idx)
    and h = set_glue next.vpos setting in
    let width = Array.fold_left add_glue_width zero_glue line in
    let st = { x = get_indent h; y = y + h; g = setting_for (get_width h) width; } in
    Array.iter (put_node st) line;
    put_par y setting get_indent get_width nodes (next :: rest);
  | _ -> ();;

(* M A I N *)

let argc = Array.length Sys.argv;;
if argc < 2 then failwith "Missing file argument"
else if argc < 3 then failwith "Missing shape argument";;

(* initializing graphical device *)
(* let width, height = Grdev.open_dev " 1000x700";; *)
let width, height = Grdev.open_dev " 1300x970";;
Graphics.set_window_title "Dyntyp";;
Graphics.auto_synchronize true;;

(* drawing (mostly) *)
let state = { x = 0; y = 100 * pt; g = Rigid; };;
let cmr10 = make_font "cmr10" 200.0;;
(* String.iter (inspect cmr10) "abcdefghijklmnopqrstuvwxyz"; *)
(* String.iter (put_char state cmr10) "ahoj";; *)
(* Grdev.fill_rect 0 100 width 5;; *)

(*
 * reading the file given on the command line
 * returning the corresponding list of nodes
 *)
let text = read_file cmr10 Sys.argv.(1);;

(* Array.iter (function c -> print_node c; print_newline ()) text;; *)
(* Array.iter (put_node state) text;; *)

let make_show indent hsize desired top height vsize vert_step = 
  let my_plot top indent hsize vsize =
    Graphics.set_color Graphics.red;
    let t = Misc.round (float top /. fpt) in
    let y = ref t
    and h = Graphics.size_y () in
    while !y * pt <= top + vsize do
      let dy = ((!y - t) * pt) in
      let x = Misc.round ((float (hsize dy) +. float (indent dy)) /. fpt) in
      Graphics.plot x (h - !y);
      Graphics.plot (x + 1) (h - !y);
      Graphics.plot (x + 2) (h - !y);
      y := succ !y
    done in
  let par, setting = best_par text hsize height (break_text desired vert_step text) in
  put_par top setting indent hsize text par;
  my_plot top indent hsize vsize;;

let top = 0;;
let vsize = (Graphics.size_y () - 12) * pt;;
let indent d = 0;;

let filling, height =
  if argc > 3 then
    let arg = int_of_string Sys.argv.(3) in
    let arg = max 0 (min arg 100) in
    (pt, vsize / 100 * arg)
  else (0, vsize);;

let vert_step _ = { zero_glue with nat = 36 * pt; fil = filling; };;
(* let height = vsize;; *)

match Sys.argv.(2).[0] with
  | '-' ->
    let hsize d = 850 * pt in
    let desired g = (hsize 0, hsize 0) in
    make_show indent hsize desired top height vsize vert_step
  | 'a' ->
    let hsize d = 700 * pt + d / 3 in
    let desired g = 
      let lower, upper = glue_bounds g vsize in
      (hsize lower, hsize upper) in
    make_show indent hsize desired top height vsize vert_step
  | 'b' ->
    let hsize d = 990 * pt - d / 3 in
    let desired g = 
      let lower, upper = glue_bounds g vsize in
      (hsize upper, hsize lower) in
    make_show indent hsize desired top height vsize vert_step
  | 'c' ->
    let hsize d =
      let r = float vsize /. 2.0 in
      let y = float d -. r in
      600 * pt + Misc.round (sqrt (r *. r -. y *. y)) in
    let desired g =
      let r = vsize / 2 in
      let lower, upper = glue_bounds g vsize in
      if upper < r then (hsize lower, hsize upper)
      else if lower > r then (hsize upper, hsize lower)
      else (min (hsize lower) (hsize upper), hsize r) in
    make_show indent hsize desired top height vsize vert_step
  | 'd' ->
    let width = 850 * pt
    and amp = 50 * pt
    and pi = 3.1415926535 in
    let hsize d =
      width + Misc.round (float amp *. sin (float (6 * d) *. pi /. float vsize)) in
    let desired g =
      (width - amp, width + amp) in
    make_show indent hsize desired top height vsize vert_step
  | 'e' ->
    let fr = float vsize /. 2.0 in
    let r = Misc.round fr in
    let half d =
      let fy = float d -. fr in
      Misc.round (sqrt (fr *. fr -. fy *. fy)) in
    let indent d = r - half d in
    let hsize d = 2 * half d in
    let desired g =
      let r = vsize / 2 in
      let lower, upper = glue_bounds g vsize in
      if upper < r then (hsize lower, hsize upper)
      else if lower > r then (hsize upper, hsize lower)
      else (min (hsize lower) (hsize upper), hsize r) in
    make_show indent hsize desired top height vsize vert_step
  | _ ->
    prerr_string "Invalid shape argument, use:\n";
    prerr_string "-:\tconstant width\n";
    prerr_string "a:\tincreasing slope\n";
    prerr_string "b:\tdecreasing slope\n";
    prerr_string "c:\thalf circle\n";
    prerr_string "d:\tsinusoid\n";
    prerr_string "e:\tcircle\n"
;;

(* waiting for a key *)

let rec wait_for_key () =
  match Grdev.wait_event () with
    | Grdev.Key c -> c
    | _ -> wait_for_key ()
;;

prerr_endline "Press a key";;
wait_for_key ();;
Grdev.close_dev ();;
